#!/usr/bin/env python
# Copyright (c) 2003-2006 ActiveState Software Inc.
#
# The MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify,
# merge, publish, distribute, sublicense, and/or sell copies of the
# Software, and to permit persons to whom the Software is furnished
# to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
#
# Authors:
#    Shane Caraveo <ShaneC@ActiveState.com>
#    Trent Mick <TrentM@ActiveState.com>

"""
    pydbgp -d localhost:9000 script.py [args]
    
    -d hostname:port  to debug a script
    -k ide_key        a IDE key used with proxies
    -p script.py      preload and execute script before debugging
    -m modules        comma deliminated list of modules to ignore
                      during debug session
    -i interactive    start debugger in interactive mode
                      if this is used in combination with a script, then
                      interactive mode will be entered when the script has
                      completed debugging
    -n                run without debugging.  This will start the debugger
                      if there is an exception.  It can also be used in
                      combination with -i to start the interactive shell when
                      the script has finished running.
    -r                Do not redirect stdin to the IDE
    -l log_level      Logging levels from the logging module:
                        CRITICAL
                        ERROR
                        WARN
                        INFO
                        DEBUG

"""
    
__version__ = (1, 1, 0)
__revision__ = "$Revision: #1 $ $Change: 118727 $"


import sys
import os

# Alternate environment variable for specifying dbgp location.  This
# allows python -E to work even if dbgp is not installed under
# site-packages
if os.environ.has_key("PYDBGP_PATH"):
    sys.path.insert(0, os.environ['PYDBGP_PATH'])

import getopt
import socket
import types

def _get_dbgp_client_pythonlib_path():
    """Find the DBGP Python client library in the common install
    configuration. Returns None if it could not be found.
    """
    from os.path import dirname, join, abspath, exists
    try:
        this_dir = dirname(abspath(__file__))
    except NameError:
        this_dir = dirname(abspath(sys.argv[0]))
    candidate_paths = [
        dirname(this_dir), # Komodo source tree layout
        join(dirname(this_dir), "pythonlib"),
    ]
    for candidate_path in candidate_paths:
        landmark = join(candidate_path, "dbgp", "__init__.py")
        if exists(landmark):
            return candidate_path

_p = (not hasattr(sys, "frozen") and _get_dbgp_client_pythonlib_path() or None)
if _p: sys.path.insert(0, _p)
try:
    import dbgp.client
    from dbgp.client import log, h_main
    from dbgp.common import *
    try:
        import logging
    except ImportError:
        from dbgp import _logging as logging
finally:
    if _p: del sys.path[0]



class IOStream:
    def __init__(self, origStream, encoding):
        self.__dict__['_origStream'] = origStream
        self.__dict__['_encoding'] = encoding
    
    def write(self, s):
        try:
            if type(s)==types.UnicodeType:
                s = s.encode(self._encoding)
        except:
            pass
        self._origStream.write(s)

    def writelines(self, lines):
        text = ''.join(lines)
        self.write(text)
        
    def __getattr__(self, attr):
        if self.__dict__.has_key(attr):
            return getattr(self,attr)
        return getattr(self._origStream, attr)

def _fixencoding():
    """If we're not run from a tty, force stdout to an encoding defined
    in LANG or to mbcs.  This is required to make python properly output
    unicode output, otherwise it just spits out an exception."""
    # based on logic found in Py_Initialize in pythonrun.c
    import locale
    codeset = locale.getdefaultlocale()[1]
    if codeset:
        try:
            import codecs
            secret_decoder_ring = codecs.lookup(codeset)
        except LookupError, e:
            if sys.platform.startswith('win'):
                codeset = 'mbcs'
            else:
                codeset = 'UTF-8'
        if not hasattr(sys.stdout, "isatty") or not sys.stdout.isatty():
            sys.stdout = IOStream(sys.stdout, codeset)
        if not hasattr(sys.stderr, "isatty") or not sys.stderr.isatty():
            sys.stderr = IOStream(sys.stdout, codeset)

def main(argv):
    logLevel = logging.WARN
    configureLogging(log, logLevel)
    _fixencoding()
    try:
        optlist, args = getopt.getopt(argv[1:], 'hVd:k:l:p:m:inr',
            ['help', 'version', 'debug_port',
             'key', 'log_level', 'preload', 'modules',
             'interactive', 'nodebug', 'nostdin'])
    except getopt.GetoptError, msg:
        sys.stderr.write("pydbgp: error: %s\n" % str(msg))
        sys.stderr.write("See 'pydbgp --help'.\n")
        return 1

    import locale
    codeset = locale.getdefaultlocale()[1]
    idekey = getenv('USER', getenv('USERNAME', ''))
    try:
        if codeset:
            idekey = idekey.decode(codeset)
        else:
            idekey = idekey.decode()
    except (UnicodeDecodeError, LookupError), e:
        log.warn("unable to decode idekey %r"%idekey)
        pass # nothing we can do if defaultlocale is wrong
    host = '127.0.0.1'
    port = 9000
    preloadScript = None
    ignoreModules = []
    interactive = 0
    nodebug = 0
    redirect = 1
    for opt, optarg in optlist:
        if optarg:
            try:
                if codeset:
                    optarg = optarg.decode(codeset)
                else:
                    optarg = optarg.decode()
            except (UnicodeDecodeError, LookupError), e:
                log.warn("unable to decode argument %s = %r"%(opt,optarg))
                pass # nothing we can do if defaultlocale is wrong
        if opt in ('-h', '--help'):
            sys.stdout.write(__doc__)
            return 0
        elif opt in ('-V', '--version'):
            import re
            kw = re.findall('\$(\w+):\s(.*?)\s\$', __revision__)
            sys.stderr.write("pydbgp Version %s %s %s %s %s\n"\
                             % ('.'.join([str(i) for i in __version__]),
                                kw[0][0], kw[0][1], kw[1][0], kw[1][1]))
            return 0
        elif opt in ('-d', '--debug_port'):
            if optarg.find(':') >= 0:
                host, port = optarg.split(':')
                port = int(port)
            else:
                host = '127.0.0.1'
                port = int(optarg)
        elif opt in ('-k', '--key'):
            idekey = optarg
        elif opt in ('-n', '--nodebug'):
            nodebug = 1
        elif opt in ('-l', '--log_level'):
            if optarg in logging._levelNames:
                logLevel = logging._levelNames[optarg]
            else:
                sys.stderr.write("pydbgp: error: Invalid log level\n")
                sys.stderr.write("See 'pydbgp --help'.\n")
                return 1
        elif opt in ('-p', '--preload'):
            preloadScript = optarg
        elif opt in ('-m', '--modules'):
            ignoreModules = optarg.split(',')
        elif opt in ('-i', '--interactive'):
            interactive = 1
        elif opt in ('-r', '--nostdin'):
            redirect = 0

    if not port:
        sys.stderr.write("pydbgp: error: IDE Port not provided\n")
        sys.stderr.write("See 'pydbgp --help'.\n")
        return 1
    
    if interactive:
        if not args:
            args = ['interactive']
            if sys.path[0] != '' and os.getcwd() not in sys.path:
                sys.path.insert(0, os.getcwd())

    if not args:
        sys.stderr.write("pydbgp: error: scriptname not provided\n")
        sys.stderr.write("See 'pydbgp --help'.\n")
        return 1
    
    # handle ~ paths
    if not interactive:
        args[0] = os.path.expanduser(args[0])
        args[0] = os.path.realpath(args[0])
        if not os.path.exists(args[0]):
            sys.stderr.write("pydbgp: error: scriptname does not exist\n")
            sys.stderr.write("See 'pydbgp --help'.\n")
            return 1
        
    if nodebug:
        dbgp.client.runWithoutDebug(args, interactive, host, port, idekey, logLevel)
    else:
        log.setLevel(logLevel)
        dbgp.client.set_thread_support(dbgp.client.backendCmd.debug_threads)
        client = dbgp.client.backendCmd(idekey, preloadScript, ignoreModules, module=h_main())
        client.stdin_enabled = redirect
        try:
            client.connect(host, port, '__main__', args)
        except socket.error, e:
            return 1
        if interactive and args[0] == 'interactive':
            cprt = 'Type "copyright", "credits" or "license" for more information.'
            sys.stdout.write("Python %s on %s\n%s\n" %
                       (sys.version, sys.platform, cprt))
            # wait until exit
            client.runInteractive()
        else:
            client.runMain(args, interactive)
    return 0

if __name__ == "__main__":
    sys.exit( main(sys.argv) )


